import collections
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, TypeVar, Union

from goeiedag._model import _Input, _Output, _PlaceholderBase, _Task, CmdArgument, InputsOutputs


K = TypeVar("K")
V = TypeVar("V")


def flatten_inputs_outputs(inputs, outputs) -> Tuple[Sequence, Sequence]:
    inputs_sequence = inputs.values() if isinstance(inputs, dict) else inputs
    outputs_sequence = outputs.values() if isinstance(outputs, dict) else outputs

    return inputs_sequence, outputs_sequence


def map_dict_values(collection: Dict[K, Any], callback: Callable[..., V]) -> Dict[K, V]:
    return {k: callback(v) for k, v in collection.items()}


def _ensure_dict(collection: Union[Sequence[V], Dict[K, V]]) -> Dict[Union[int, K], V]:
    if isinstance(collection, dict):
        # Mypy complains here about relaxing Dict[K, V] -> Dict[int | K, V]
        return collection  # type: ignore[return-value]
    else:
        return {i: v for i, v in enumerate(collection)}


def _resolve_placeholder(arg, mapping: dict) -> Sequence[str]:
    if arg.name is None:  # All inputs (outputs)
        assert arg.prefix == ""

        return list(mapping.values())
    elif arg.name == "":  # Unique input (output)
        assert len(mapping) == 1

        return [arg.prefix + str(mapping[0])]
    else:  # Specific input (output)
        assert arg.name is not None
        assert isinstance(mapping, dict)

        return [arg.prefix + str(mapping[arg.name])]


def resolve_placeholders(command: Sequence,
                         inputs: Dict[Union[int, str], str],
                         outputs: Dict[Union[int, str], str]) -> Sequence[CmdArgument]:
    command_expanded: List[CmdArgument] = []

    for arg in command:
        if isinstance(arg, _Input):
            command_expanded += _resolve_placeholder(arg, inputs)
        elif isinstance(arg, _Output):
            command_expanded += _resolve_placeholder(arg, outputs)
        else:
            command_expanded.append(arg)

    return command_expanded


class CommandGraph:
    aliases: Dict[str, Sequence[Path]]
    tasks: List[_Task]

    def __init__(self):
        self.aliases = {}
        self.tasks = []

    def add(
            self,
            command: Sequence[CmdArgument | _PlaceholderBase],
            *,
            inputs: Sequence[Path | str] | Dict[str, Path | str],
            outputs: Sequence[Path | str] | Dict[str, Path | str]
    ) -> None:
        self.tasks.append(
            _Task(
                command=command,
                # From now on, all inputs/outpus must be Paths
                inputs=map_dict_values(_ensure_dict(inputs), Path),
                outputs=map_dict_values(_ensure_dict(outputs), Path),
            )
        )

    def add_alias(self,
                  *args: Union[Path, str],
                  name: Optional[str] = None) -> str:
        if name is None:
            name = f"_alias_{len(self.aliases)}"

        assert name not in self.aliases
        self.aliases[name] = [Path(arg) for arg in args]
        return name



from pathlib import Path
import tempfile

import goeiedag
from goeiedag import ALL_INPUTS, INPUT, OUTPUT


def test_basic():
    with tempfile.TemporaryDirectory() as dir:
        graph = goeiedag.CommandGraph()

        # Extract OS name from /etc/os-release
        graph.add(["grep", "^NAME=", INPUT, ">", OUTPUT],
                  inputs=["/etc/os-release"],
                  outputs=["os-name.txt"])
        # Get username
        graph.add(["whoami", ">", OUTPUT],
                  inputs=[],
                  outputs=["username.txt"])
        # Glue together to produce output
        graph.add(["cat", ALL_INPUTS, ">", OUTPUT.result],
                  inputs=["os-name.txt", "username.txt"],
                  outputs=dict(result="result.txt"))

        goeiedag.build_all(graph, Path(dir))

        assert (Path(dir) / "result.txt").stat().st_size > 0


def test_escape():
    with tempfile.TemporaryDirectory() as dir:
        graph = goeiedag.CommandGraph()

        graph.add(["echo", "Hello", ">", OUTPUT], inputs=[], outputs=[">weirdfilename"])

        goeiedag.build_all(graph, Path(dir))

        assert (Path(dir) / ">weirdfilename").stat().st_size > 0


def test_alias():
    with tempfile.TemporaryDirectory() as dir:
        graph = goeiedag.CommandGraph()

        graph.add(["echo", "Hello", ">", OUTPUT], inputs=[], outputs=["file1"])
        graph.add(["echo", "Hello", ">", OUTPUT], inputs=[], outputs=["file2"])
        graph.add(["echo", "Hello", ">", OUTPUT], inputs=[], outputs=["file3"])
        graph.add_alias("file1", Path("file3"), name="files1_3")

        goeiedag.build_targets(graph, Path(dir), ["files1_3"])

        assert (Path(dir) / "file1").exists()
        assert not (Path(dir) / "file2").exists()
        assert (Path(dir) / "file3").exists()


def test_outputs_dict():
    with tempfile.TemporaryDirectory() as dir:
        graph = goeiedag.CommandGraph()

        graph.add(["touch", OUTPUT.foo], inputs=[], outputs=dict(foo="file1"))
        graph.add(["touch", OUTPUT.foo], inputs=[], outputs=dict(foo="file2"))

        goeiedag.build_all(graph, Path(dir))



import logging
from pathlib import Path
import re
import shlex
import subprocess
import time
from typing import List, Optional, Sequence

import ninja

from ._graph import CommandGraph, map_dict_values, resolve_placeholders
from ._model import BuildFailure, CmdArgument


logger = logging.getLogger(__name__)


def _generate_rule_name(command: Sequence[CmdArgument], i: int) -> str:
    # try to extract something that represents the command intuitively (just for ninjafile debugging)
    blurb = " ".join(arg.name if isinstance(arg, Path) else str(arg) for arg in command)[:40]

    sanitized = _sanitize_rule_name(blurb)

    if not len(sanitized):
        sanitized = "rule"

    return f"{sanitized}_{i}"


# https://stackoverflow.com/a/23532381
_full_pattern = re.compile("[^a-zA-Z0-9_]|_")


def _sanitize_rule_name(string: str) -> str:
    return re.sub(_full_pattern, "_", string)


def write_ninja_file(g: CommandGraph, output):
    writer = ninja.Writer(output)

    rule_names: List[str] = []

    # flatten everything
    flat_tasks = []
    for task in g.tasks:
        # Inputs/outputs must be shell-escaped before being substituted into the command.
        # This way, literal tokens in the command will be preserved, allowing use of shell features like output
        # redirections, but inserted input/output names will be sanitized.
        inputs_shellesc = map_dict_values(task.inputs, lambda path: shlex.quote(str(path)))
        outputs_shellesc = map_dict_values(task.outputs, lambda path: shlex.quote(str(path)))

        command = resolve_placeholders(task.command, inputs_shellesc, outputs_shellesc)

        flat_tasks.append((task.inputs, task.outputs, command))

    # generate rules
    for inputs, outputs, command in flat_tasks:
        rule_name = _generate_rule_name(command, len(rule_names))
        rule_names.append(rule_name)

        writer.rule(
            name=rule_name,
            command=" ".join(ninja.escape(str(arg)) for arg in command),
        )
        writer.newline()

    # emit build statements
    for i, (inputs, outputs, command) in enumerate(flat_tasks):
        writer.build(
            rule=rule_names[i],
            inputs=[str(i) for i in inputs.values()],
            outputs=[str(i) for i in outputs.values()],
        )
        writer.newline()

        if len(outputs):
            writer.default([str(i) for i in outputs.values()])
            writer.newline()

    # emit Phony statements for aliases
    for name, inputs_list in g.aliases.items():
        writer.build(rule="phony", inputs=[str(i) for i in inputs_list], outputs=[name])

    writer.close()
    del writer


def build_targets(g: CommandGraph, build_dir: Path, targets: Optional[Sequence[CmdArgument]], cwd: Optional[Path] = None):
    # targets = outputs | aliases
    if targets and not len(targets):
        return  # nothing to do

    build_dir.mkdir(exist_ok=True)
    ninjafile_path = build_dir / "build.ninja"

    pre = time.time()

    with open(ninjafile_path, "wt") as output:
        write_ninja_file(g, output)

    post = time.time()

    logger.info("write_ninja_file took %d msec", int((post - pre) * 1000))

    pre = time.time()
    extra_arguments = [str(x) for x in targets] if targets else []

    try:
        subprocess.check_call([Path(ninja.BIN_DIR) / "ninja", "-f", ninjafile_path.absolute()] + extra_arguments, cwd=cwd or build_dir)
    except subprocess.CalledProcessError as ex:
        raise BuildFailure(f"Ninja build returned error code {ex.returncode}") from None
    finally:
        post = time.time()

        logger.info("Ninja build took %d msec", int((post - pre) * 1000))

def build_all(g: CommandGraph, build_dir: Path, cwd: Optional[Path] = None):
    return build_targets(g, build_dir, targets=None, cwd=cwd)



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Sequence, Union


@dataclass
class _PlaceholderBase:
    name: Optional[str]
    prefix: str = ""

    def __getattr__(self, name: str) -> "_PlaceholderBase":
        assert self.name == ""
        return self.__class__(name)

    def __radd__(self, other: str) -> "_PlaceholderBase":
        if not isinstance(other, str):
            raise TypeError()

        return self.__class__(name=self.name, prefix=other + self.prefix)


class _Input(_PlaceholderBase):
    pass


class _Output(_PlaceholderBase):
    pass


CmdArgument = Path | str
InputsOutputs = Dict[Union[int, str], Path]
ALL_INPUTS = _Input(None)
ALL_OUTPUTS = _Output(None)
INPUT = _Input("")  # unique input (asserted)
OUTPUT = _Output("")   # unique output (asserted)


@dataclass
class _Task:
    command: Sequence[Union[CmdArgument, _PlaceholderBase]]
    inputs: InputsOutputs
    outputs: InputsOutputs


# Deliberately not named BuildError, because it represents a non-specific failure of the build as a whole
class BuildFailure(Exception):
    pass
